"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * SonarTS
 * Copyright (C) 2017-2019 SonarSource SA
 * mailto:info AT sonarsource DOT com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
const tslint = require("tslint");
const ts = require("typescript");
const nodes_1 = require("../utils/nodes");
const sonarAnalysis_1 = require("../utils/sonarAnalysis");
class Rule extends tslint.Rules.TypedRule {
    applyWithProgram(sourceFile, program) {
        return new Visitor(this.getOptions().ruleName, program).visit(sourceFile).getIssues();
    }
}
Rule.metadata = {
    ruleName: "no-useless-cast",
    description: "Redundant casts and not-null assertions should be avoided",
    optionsDescription: "",
    options: null,
    rspecKey: "RSPEC-4325",
    type: "functionality",
    typescriptOnly: false,
};
Rule.MESSAGE_CAST = "Remove this unnecessary cast.";
Rule.MESSAGE_ASSERTION = "Remove this unnecessary not-null assertion.";
exports.Rule = Rule;
class Visitor extends sonarAnalysis_1.TypedSonarRuleVisitor {
    // for some reason there is no specific "visit..." method for AsExpression
    visitNode(node) {
        if (nodes_1.is(node, ts.SyntaxKind.AsExpression, ts.SyntaxKind.TypeAssertionExpression)) {
            this.checkTypeCasting(node);
        }
        super.visitNode(node);
    }
    visitNonNullExpression(node) {
        const typeAtLocation = this.program.getTypeChecker().getTypeAtLocation(node.expression);
        if (typeAtLocation.getFlags() & ts.TypeFlags.Union) {
            const { types } = typeAtLocation;
            if (!types.some(isUndefinedOrNull)) {
                this.addIssue(node, Rule.MESSAGE_ASSERTION);
            }
        }
        else if (!isUndefinedOrNull(typeAtLocation)) {
            this.addIssue(node, Rule.MESSAGE_ASSERTION);
        }
        super.visitNonNullExpression(node);
    }
    checkTypeCasting(assertionExpression) {
        const { expression, type } = assertionExpression;
        const actualExpressionType = this.program.getTypeChecker().getTypeAtLocation(expression);
        const typeToCast = this.program.getTypeChecker().getTypeFromTypeNode(type);
        const compatibleTypes = [actualExpressionType, ...getBaseTypes(actualExpressionType)];
        if (nodes_1.is(expression, ts.SyntaxKind.StringLiteral, ts.SyntaxKind.NumericLiteral) &&
            nodes_1.is(type, ts.SyntaxKind.LiteralType) &&
            expression.kind === type.literal.kind) {
            // allow singleton types: '<"a"> "a";' or '0 as 0;'
            return;
        }
        if ((nodes_1.is(expression, ts.SyntaxKind.NumericLiteral) && nodes_1.is(type, ts.SyntaxKind.NumberKeyword)) ||
            (nodes_1.is(expression, ts.SyntaxKind.StringLiteral) && nodes_1.is(type, ts.SyntaxKind.StringKeyword)) ||
            compatibleTypes.includes(typeToCast)) {
            this.addIssue(assertionExpression, Rule.MESSAGE_CAST);
        }
    }
}
function isUndefinedOrNull(type) {
    return !!(type.flags & ts.TypeFlags.Undefined) || !!(type.flags & ts.TypeFlags.Null);
}
function getBaseTypes(type) {
    const baseTypes = type.getBaseTypes();
    if (baseTypes) {
        const result = [...baseTypes];
        baseTypes.forEach(baseType => result.push(...getBaseTypes(baseType)));
        return result;
    }
    else {
        return [];
    }
}
//# sourceMappingURL=noUselessCastRule.js.map